<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\BaseExt\Metadata\Driver;

use Metadata\Driver\DriverInterface;
use Doctrine\Common\Annotations\Reader;
use IMIA\BaseExt\Metadata\ClassMetadata;
use IMIA\BaseExt\Metadata\PropertyMetadata;

/**
 * @package     imia_base_ext
 * @subpackage  Metadata\Driver
 * @author      David Frerich <d.frerich@imia.de>
 */
class AnnotationDriver implements DriverInterface
{
    /**
     * @var \Doctrine\Common\Annotations\Reader
     */
    private $reader;

    /**
     * @var string
     */
    private $mergeType;

    /**
     * @var array
     */
    private $annotations = array(
        'classes'       => array(
            'sqlTable'                  => 'IMIA\BaseExt\Annotation\SQL\Table',
            'sqlIndex'                  => 'IMIA\BaseExt\Annotation\SQL\Index',
            'tcaTable'                  => 'IMIA\BaseExt\Annotation\TCA\Table',
            'tcaAddPaletteToTypes'      => 'IMIA\BaseExt\Annotation\TCA\AddPaletteToTypes',
            'tcaAddPalettesToTypes'     => 'IMIA\BaseExt\Annotation\TCA\AddPalettesToTypes',
        ),
        'properties'        => array(
            'sqlColumn'     => 'IMIA\BaseExt\Annotation\SQL\Column',
            'tcaColumn'     => array(
                'IMIA\BaseExt\Annotation\TCA\Categories',
                'IMIA\BaseExt\Annotation\TCA\Check',
                'IMIA\BaseExt\Annotation\TCA\Column',
                'IMIA\BaseExt\Annotation\TCA\Files',
                'IMIA\BaseExt\Annotation\TCA\FilesImage',
                'IMIA\BaseExt\Annotation\TCA\FilesPDF',
                'IMIA\BaseExt\Annotation\TCA\Flex',
                'IMIA\BaseExt\Annotation\TCA\Group',
                'IMIA\BaseExt\Annotation\TCA\Inline',
                'IMIA\BaseExt\Annotation\TCA\Input',
                'IMIA\BaseExt\Annotation\TCA\InputLink',
                'IMIA\BaseExt\Annotation\TCA\InputNumber',
                'IMIA\BaseExt\Annotation\TCA\InputColorpicker',
                'IMIA\BaseExt\Annotation\TCA\InputDate',
                'IMIA\BaseExt\Annotation\TCA\InputDateTime',
                'IMIA\BaseExt\Annotation\TCA\InputSlider',
                'IMIA\BaseExt\Annotation\TCA\None',
                'IMIA\BaseExt\Annotation\TCA\Passthrough',
                'IMIA\BaseExt\Annotation\TCA\Radio',
                'IMIA\BaseExt\Annotation\TCA\Select',
                'IMIA\BaseExt\Annotation\TCA\Text',
                'IMIA\BaseExt\Annotation\TCA\TextRTE',
                'IMIA\BaseExt\Annotation\TCA\User',

                /* @deprecated since version 1.1.14 */
                'IMIA\BaseExt\Annotation\TCA\Colorpicker',
                'IMIA\BaseExt\Annotation\TCA\DateTime',
                'IMIA\BaseExt\Annotation\TCA\Number',
                'IMIA\BaseExt\Annotation\TCA\ImageFiles',
            ),
            'tcaSearchField'     => 'IMIA\BaseExt\Annotation\TCA\SearchField',
            'tcaAddToTypes'      => 'IMIA\BaseExt\Annotation\TCA\AddToTypes',
        ),
    );

    /**
     * @param \Doctrine\Common\Annotations\Reader $reader
     * @param boolean $mergeAll
     */
    public function __construct(Reader $reader, $mergeType = null)
    {
        $this->reader = $reader;
        $this->mergeType = $mergeType;

        $this->loadAnnotations();
    }

    /**
     * @param \ReflectionClass $class
     * @return \IMIA\BaseExt\Metadata\ClassMetadata
     */
    public function loadMetadataForClass(\ReflectionClass $class)
    {
        try {
            $classMetadata = new ClassMetadata($class->getName());
            $classMetadata->mergeType = $this->mergeType;

            $this->loadClassMetadata($classMetadata, $class);
            $this->loadPropertyMetadata($classMetadata, $class);
        } catch(\Exception $e) {
            echo '<pre>';
            echo '<h2>Annotation Exception</h2>';
            print_r($e->getMessage());
            exit;
        }

        return $classMetadata;
    }

    /**
     * @param IMIA\BaseExt\Metadata\ClassMetadata $classMetadata
     * @param \ReflectionClass $reflectionClass
     * @return \IMIA\BaseExt\Metadata\ClassMetadata
     */
    private function loadClassMetadata(&$classMetadata, $reflectionClass)
    {
        foreach ($this->annotations['classes'] as $field => $annotationClasses) {
            if (is_string($annotationClasses)) {
                $this->setFieldFromAnnotation($classMetadata, $field, $reflectionClass, $annotationClasses);
            } else {
                foreach ($annotationClasses as $annotationClass) {
                    if ($this->setFieldFromAnnotation($classMetadata, $field, $reflectionClass, $annotationClass)) {
                        break;
                    }
                }
            }
        }

        return $classMetadata;
    }

    /**
     * @param \IMIA\BaseExt\Metadata\ClassMetadata $classMetadata
     * @param \ReflectionClass $reflectionClass
     * @return \IMIA\BaseExt\Metadata\ClassMetadata
     */
    private function loadPropertyMetadata(&$classMetadata, $reflectionClass)
    {
        foreach ($reflectionClass->getProperties() as $reflectionProperty) {
            $hasOwnProperty = true;
            if ($reflectionClass->getParentClass() && $reflectionClass->getParentClass()->hasProperty($reflectionProperty->getName())) {
                if ($reflectionClass->getParentClass()->getProperty($reflectionProperty->getName())->getDocComment() == $reflectionProperty->getDocComment()) {
                    $hasOwnProperty = false;
                }
            }

            if ($hasOwnProperty) {
                $propertyMetadata = new PropertyMetadata($reflectionClass->getName(), $reflectionProperty->getName());
                foreach ($this->annotations['properties'] as $field => $annotationClasses) {
                    if (is_string($annotationClasses)) {
                        $this->setFieldPropertyFromAnnotation($propertyMetadata, $field, $reflectionProperty, $annotationClasses);
                    } else {
                        foreach ($annotationClasses as $annotationClass) {
                            if ($this->setFieldPropertyFromAnnotation($propertyMetadata, $field, $reflectionProperty, $annotationClass)) {
                                break;
                            }
                        }
                    }
                }

                $classMetadata->addPropertyMetadata($propertyMetadata);
            }
        }

        return $classMetadata;
    }

    /**
     * @param \IMIA\BaseExt\Metadata\ClassMetadata $classMetadata
     * @param string $field
     * @param \ReflectionClass $reflectionClass
     * @param string|array $annotationClass
     * @return boolean
     */
    private function setFieldFromAnnotation(&$classMetadata, $field, $reflectionClass, $annotationClass)
    {
        $success = false;
        $annotation = $this->reader->getClassAnnotation(
            $reflectionClass,
            $annotationClass
        );

        if (null !== $annotation) {
            $classMetadata->$field = $annotation;
            $success = true;
        }

        return $success;
    }

    /**
     * @param \IMIA\BaseExt\Metadata\PropertyMetadata $propertyMetadata
     * @param string $field
     * @param \ReflectionClass $reflectionClass
     * @param string|array $annotationClass
     * @return boolean
     */
    private function setFieldPropertyFromAnnotation(&$propertyMetadata, $field, $reflectionProperty, $annotationClass)
    {
        $success = false;
        $annotation = $this->reader->getPropertyAnnotation(
            $reflectionProperty,
            $annotationClass
        );

        if (null !== $annotation) {
            $propertyMetadata->$field = $annotation;
            $success = true;
        }

        return $success;
    }

    /**
     */
    private function loadAnnotations()
    {
        foreach ($this->loadAnnotationClasses() as $annotationFile) {
            require_once($annotationFile);
        }
    }

    /**
     * @return array
     */
    private function loadAnnotationClasses()
    {
        return (array)@include(\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::
            extPath('imia_base_ext', 'ext_autoload_annotations.php'));
    }
}